from django.conf import settings

ESI_SSO_CLIENT_ID = getattr(settings, 'ESI_SSO_CLIENT_ID', None)
"""Client ID of your Eve Online SSO app. REQUIRED."""

ESI_SSO_CLIENT_SECRET = getattr(settings, 'ESI_SSO_CLIENT_SECRET', None)
"""Client Secret of your Eve Online SSO app. REQUIRED."""

ESI_SSO_CALLBACK_URL = getattr(settings, 'ESI_SSO_CALLBACK_URL', None)
"""Callback for your Eve Online SSO app. REQUIRED."""

# ESI_SSO_CLIENT_ID, ESI_SSO_CLIENT_SECRET ESI_SSO_CALLBACK_URL
# are required for SSO to function.
# Can be left blank if settings.DEBUG is set to True

ESI_API_DATASOURCE = getattr(settings, 'ESI_API_DATASOURCE', 'tranquility')
"""Change these to switch to Singularity."""

ESI_API_VERSION = getattr(settings, 'ESI_API_VERSION', 'latest')
"""Change this to access different revisions of the ESI API by default"""

ESI_ALWAYS_CREATE_TOKEN = getattr(settings, 'ESI_ALWAYS_CREATE_TOKEN', False)
"""Enable to force new token creation every callback."""

ESI_CACHE_RESPONSE = getattr(settings, 'ESI_CACHE_RESPONSE', True)
"""Disable to stop caching endpoint responses."""

ESI_INFO_LOGGING_ENABLED = getattr(settings, 'ESI_INFO_LOGGING_ENABLED', False)
"""Enable/disable verbose info logging."""

ESI_LOG_LEVEL_LIBRARIES = str(getattr(settings, 'ESI_LOG_LEVEL_LIBRARIES', 'INFO'))
"""Set log level for libraries like bravado and urlib3."""


ESI_USER_CONTACT_EMAIL = getattr(settings, 'ESI_USER_CONTACT_EMAIL', None)
"""Contact email address of server owner.

This will be included in the User-Agent header of every request.
"""

ESI_CONNECTION_POOL_MAXSIZE = getattr(settings, 'ESI_CONNECTION_POOL_MAXSIZE', 10)
"""Max size of the connection pool.

Increase this setting if you hav more parallel
threads connected to ESI at the same time.
"""

ESI_CONNECTION_ERROR_MAX_RETRIES = getattr(
    settings, 'ESI_CONNECTION_ERROR_MAX_RETRIES', 3
)
"""Max retries on failed connections."""

ESI_SERVER_ERROR_MAX_RETRIES = getattr(settings, 'ESI_SERVER_ERROR_MAX_RETRIES', 3)
"""Max retries on server errors."""

ESI_SERVER_ERROR_BACKOFF_FACTOR = getattr(
    settings, 'ESI_SERVER_ERROR_BACKOFF_FACTOR', 0.2
)
"""Backoff factor for retries on server error."""

ESI_REQUESTS_CONNECT_TIMEOUT = getattr(settings, 'ESI_REQUESTS_CONNECT_TIMEOUT', 5)
"""Default connection timeouts for all requests to ESI.

Can temporarily overwritten with by passing ``timeout`` with ``result()``
"""

ESI_REQUESTS_READ_TIMEOUT = getattr(settings, 'ESI_REQUESTS_READ_TIMEOUT', 30)
"""Default read timeouts for all requests to ESI.

Can temporarily overwritten with by passing ``timeout`` with ``result()``
"""

# These probably won't ever change. Override if needed.
ESI_API_URL = getattr(settings, 'ESI_API_URL', 'https://esi.evetech.net/')
ESI_OAUTH_URL = getattr(
    settings, 'ESI_SSO_BASE_URL', 'https://login.eveonline.com/v2/oauth'
)
ESI_OAUTH_LOGIN_URL = getattr(
    settings, 'ESI_SSO_LOGIN_URL', ESI_OAUTH_URL + "/authorize/"
)
ESI_TOKEN_URL = getattr(settings, 'ESI_CODE_EXCHANGE_URL', ESI_OAUTH_URL + "/token")

ESI_TOKEN_VERIFY_URL = getattr(
    settings, 'ESI_TOKEN_EXCHANGE_URL', ESI_OAUTH_URL + "/verify"
)  # deprecated
ESI_TOKEN_JWK_SET_URL = "https://login.eveonline.com/oauth/jwks"

ESI_TOKEN_VALID_DURATION = int(getattr(settings, 'ESI_TOKEN_VALID_DURATION', 1170))
ESI_SPEC_CACHE_DURATION = int(getattr(settings, 'ESI_SPEC_CACHE_DURATION', 3600))

# list of all official language codes supported by ESI
ESI_LANGUAGES = getattr(settings, 'ESI_LANGUAGES', [
    'de', 'en-us', 'fr', 'ja', 'ru', 'zh', 'ko'
])



from datetime import datetime
from hashlib import md5
import json
import logging
from time import sleep
from urllib import parse as urlparse
from typing import Any, Union, Tuple

from bravado.client import SwaggerClient
from bravado import requests_client
from bravado.exception import (
    HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
)
from bravado_core.response import IncomingResponse
from bravado.swagger_model import Loader
from bravado.http_future import HttpFuture
from bravado_core.spec import Spec, CONFIG_DEFAULTS
from requests.adapters import HTTPAdapter

from django.core.cache import cache

from .errors import TokenExpiredError
from . import app_settings, __version__, __title__


logger = logging.getLogger(__name__)

_LIBRARIES_LOG_LEVEL = logging.getLevelName(app_settings.ESI_LOG_LEVEL_LIBRARIES)
logging.getLogger('swagger_spec_validator').setLevel(_LIBRARIES_LOG_LEVEL)
logging.getLogger('bravado_core').setLevel(_LIBRARIES_LOG_LEVEL)
logging.getLogger('urllib3').setLevel(_LIBRARIES_LOG_LEVEL)
logging.getLogger('bravado').setLevel(_LIBRARIES_LOG_LEVEL)

SPEC_CONFIG = {'use_models': False}
RETRY_SLEEP_SECS = 1


class CachingHttpFuture(HttpFuture):
    """Extended wrapper for a FutureAdapter that returns a HTTP response
    and also supports caching.

    This class contains the response for an ESI request with an ESI client.
    """
    def _cache_key(self) -> str:
        """Generate the key name used to cache responses."""
        request = self.future.request
        str_hash = md5(
            (
                request.method
                + request.url
                + str(request.params)
                + str(request.data)
                + str(request.json)
            ).encode('utf-8')
        ).hexdigest()
        return 'esi_%s' % str_hash

    @staticmethod
    def _time_to_expiry(expires):
        """Determine the seconds until a HTTP header "Expires" timestamp.

        Args:
            expires: HTTP response "Expires" header

        Returns:
            seconds until "Expires" time
        """
        try:
            expires_dt = datetime.strptime(str(expires), '%a, %d %b %Y %H:%M:%S %Z')
            delta = expires_dt - datetime.utcnow()
            return delta.total_seconds()
        except ValueError:
            return 0

    def results(self, **kwargs) -> Union[Any, Tuple[Any, IncomingResponse]]:
        """Executes the request and returns the response from ESI for the current
        route. Response will include all pages if there are more available.

        Args:
            timeout: (optional) timeout for ESI request in seconds, overwrites default

        Returns:
            Response from endpoint or a tuple with response from endpoint \
            and an incoming response object containing additional meta data \
            including the HTTP response headers
        """
        results = list()
        headers = None
        # preserve original value
        _also_return_response = self.request_config.also_return_response
        # override to always get the raw response for expiry header
        self.request_config.also_return_response = True

        if "page" in self.operation.params:
            current_page = 1
            total_pages = 1

            # loop all pages and add data to output array
            while current_page <= total_pages:
                self.future.request.params["page"] = current_page
                # will use cache if applicable
                result, headers = self.result(**kwargs)
                total_pages = int(headers.headers['X-Pages'])
                # append to results list to be seamless to the client
                results += result
                current_page += 1
        else:  # it doesn't so just return
            results, headers = self.result(**kwargs)

        # restore original value
        self.request_config.also_return_response = _also_return_response

        # obey the output
        if self.request_config.also_return_response:
            return results, headers
        else:
            return results

    def results_localized(self, languages: list = None, **kwargs) -> dict:
        """Executes the request and returns the response from ESI for all default
        languages and pages (if any).

        Args:
            languages: (optional) list of languages to return \
                instead of default languages
            timeout: (optional) timeout for ESI request in seconds, overwrites default

        Returns:
            Dict of all responses with the language code as keys.
        """
        if not languages:
            my_languages = list(app_settings.ESI_LANGUAGES)
        else:
            my_languages = []
            for lang in dict.fromkeys(languages):
                if lang not in app_settings.ESI_LANGUAGES:
                    raise ValueError('Invalid language code: %s' % lang)
                my_languages.append(lang)

        return {
            language: self.results(language=language, **kwargs)
            for language in my_languages
        }

    def result(self, **kwargs) -> Union[Any, Tuple[Any, IncomingResponse]]:
        """Executes the request and returns the response from ESI. Response will
        include the requested / first page only if there are more pages available.

        Args:
            timeout: (optional) timeout for ESI request in seconds, overwrites default
            retries: (optional) max number of retries, overwrites default

        Returns:
            Response from endpoint or a tuple with response from endpoint \
            and an incoming response object containing additional meta data \
            including the HTTP response headers
        """
        if 'language' in kwargs.keys():
            # this parameter is not supported by bravado, so we can't pass it on
            self.future.request.params['language'] = str(kwargs.pop('language'))

        if 'retries' in kwargs.keys():
            max_retries = int(kwargs.pop('retries'))
        else:
            max_retries = int(app_settings.ESI_SERVER_ERROR_MAX_RETRIES)

        max_retries = max(0, max_retries)

        if 'timeout' not in kwargs:
            kwargs['timeout'] = (
                app_settings.ESI_REQUESTS_CONNECT_TIMEOUT,
                app_settings.ESI_REQUESTS_READ_TIMEOUT
            )

        if (
            app_settings.ESI_CACHE_RESPONSE
            and self.future.request.method == 'GET'
            and self.operation is not None
        ):
            result = None
            response = None
            cache_key = self._cache_key()
            try:
                cached = cache.get(cache_key)
            except Exception:
                cached = None
                logger.warning(
                    "Attempt to read ESI results from cache failed", exc_info=True
                )

            if cached:
                result, response = cached
                expiry = self._time_to_expiry(str(response.headers.get('Expires')))
                if expiry < 0:
                    logger.warning(
                        "cache expired by %d seconds, Forcing expiry", expiry
                    )
                    cached = False

            if not cached:
                # preserve original value
                _also_return_response = self.request_config.also_return_response
                # override to always get the raw response for expiry header
                self.request_config.also_return_response = True

                retries = 0
                while retries <= max_retries:
                    try:
                        if app_settings.ESI_INFO_LOGGING_ENABLED:
                            params = self.future.request.params
                            logger.info(
                                'Fetching from ESI: %s%s%s',
                                self.future.request.url,
                                f' - language {params["language"]}'
                                if 'language' in params else '',
                                f' - page {params["page"]}'
                                if 'page' in params else ''
                            )
                        logger.debug(
                            'ESI request: %s - %s',
                            self.future.request.url,
                            self.future.request.params
                        )
                        logger.debug(
                            'ESI request headers: %s', self.future.request.headers
                        )
                        result, response = super().result(**kwargs)
                        logger.debug(
                            'ESI response status code: %s', response.status_code
                        )
                        logger.debug('ESI response headers: %s', response.headers)
                        logger.debug('ESI response content: %s', response.text)
                        break
                    except (
                        HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
                    ) as ex:
                        if retries < max_retries:
                            logger.warning(
                                "ESI error (Retry: %d/%d)",
                                retries,
                                max_retries,
                                exc_info=True
                            )
                            if retries > 0:
                                wait_secs = (
                                    app_settings.ESI_SERVER_ERROR_BACKOFF_FACTOR
                                    * (2 ** (retries - 1))
                                )
                                sleep(wait_secs)
                            retries += 1
                        else:
                            raise ex

                # restore original value
                self.request_config.also_return_response = _also_return_response

                if response and 'Expires' in response.headers:
                    expires = self._time_to_expiry(response.headers['Expires'])
                    if expires > 0:
                        try:
                            cache.set(cache_key, (result, response), expires)
                        except Exception:
                            logger.warning(
                                "Failed to write ESI result to cache", exc_info=True
                            )

            if self.request_config.also_return_response:
                return result, response
            else:
                return result
        else:
            return super().result(**kwargs)


requests_client.HttpFuture = CachingHttpFuture


class TokenAuthenticator(requests_client.Authenticator):
    """
    Adds the authorization header containing access token, if specified.
    Sets ESI datasource to tranquility or singularity.
    """

    def __init__(self, token=None, datasource=None):
        host = urlparse.urlsplit(app_settings.ESI_API_URL).hostname
        super().__init__(host)
        self.token = token
        self.datasource = datasource

    def apply(self, request):
        if self.token and self.token.expired:
            if self.token.can_refresh:
                self.token.refresh()
            else:
                raise TokenExpiredError()
        request.headers['Authorization'] = \
            'Bearer ' + self.token.access_token if self.token else None
        request.params['datasource'] = \
            self.datasource or app_settings.ESI_API_DATASOURCE
        return request


class RequestsClientPlus(requests_client.RequestsClient):
    """RequestsClient with ability to set the user agent header for all requests"""

    def __init__(
        self,
        ssl_verify=True,
        ssl_cert=None,
        future_adapter_class=requests_client.RequestsFutureAdapter,
        response_adapter_class=requests_client.RequestsResponseAdapter,
    ):
        super().__init__(
            ssl_verify, ssl_cert, future_adapter_class, response_adapter_class
        )
        self.user_agent = None

    def request(
        self, request_params, operation=None, request_config=None
    ) -> HttpFuture:
        if self.user_agent:
            current_headers = request_params.get("headers", dict())
            new_header = {"User-Agent": str(self.user_agent)}
            request_params["headers"] = {**current_headers, **new_header}

        return super().request(request_params, operation, request_config)


def build_cache_name(name):
    """
    Cache key name formatter
    :param name: Name of the spec dict to cache, usually version
    :return: String name for cache key
    :rtype: str
    """
    return 'esi_swaggerspec_%s' % name


def cache_spec(name, spec):
    """
    Cache the spec dict
    :param name: Version name
    :param spec: Spec dict
    :return: True if cached
    """
    return cache.set(
        build_cache_name(name), spec, app_settings.ESI_SPEC_CACHE_DURATION
    )


def build_spec_url(spec_version):
    """
    Generates the URL to swagger.json for the ESI version
    :param spec_version: Name of the swagger spec version, like latest or v4
    :return: URL to swagger.json for the requested spec version
    """
    return urlparse.urljoin(app_settings.ESI_API_URL, spec_version + '/swagger.json')


def get_spec(name, http_client=None, config=None):
    """
    :param name: Name of the revision of spec, eg latest or v4
    :param http_client: Requests client used for retrieving specs
    :param config: Spec configuration - see Spec.CONFIG_DEFAULTS
    :return: :class:`bravado_core.spec.Spec`
    """
    http_client = http_client or requests_client.RequestsClient()

    def load_spec():
        loader = Loader(http_client)
        return loader.load_spec(build_spec_url(name))

    spec_dict = cache.get_or_set(
        build_cache_name(name), load_spec, app_settings.ESI_SPEC_CACHE_DURATION
    )
    config = dict(CONFIG_DEFAULTS, **(config or {}))
    return Spec.from_dict(spec_dict, build_spec_url(name), http_client, config)


def build_spec(base_version, http_client=None, **kwargs):
    """
    Generates the Spec used to initialize a SwaggerClient,
    supporting mixed resource versions
    :param http_client: :class:`bravado.requests_client.RequestsClient`
    :param base_version: Version to base the spec on.
    Any resource without an explicit version will be this.
    :param kwargs: Explicit resource versions, by name (eg Character='v4')
    :return: :class:`bravado_core.spec.Spec`
    """
    base_spec = get_spec(base_version, http_client=http_client, config=SPEC_CONFIG)
    if kwargs:
        for resource, resource_version in kwargs.items():
            versioned_spec = get_spec(
                resource_version, http_client=http_client, config=SPEC_CONFIG
            )
            try:
                spec_resource = versioned_spec.resources[resource.capitalize()]
            except KeyError:
                raise AttributeError(
                    'Resource {} not found on API revision {}'.format(
                        resource, resource_version
                    )
                )
            base_spec.resources[resource.capitalize()] = spec_resource
    return base_spec


def read_spec(path, http_client=None):
    """
    Reads in a swagger spec file used to initialize a SwaggerClient
    :param path: String path to local swagger spec file.
    :param http_client: :class:`bravado.requests_client.RequestsClient`
    :return: :class:`bravado_core.spec.Spec`
    """
    with open(path, encoding='utf-8') as f:
        spec_dict = json.loads(f.read())

    return SwaggerClient.from_spec(
        spec_dict, http_client=http_client, config=SPEC_CONFIG
    )


def esi_client_factory(
    token=None,
    datasource: str = None,
    spec_file: str = None,
    version: str = None,
    app_info_text: str = None,
    **kwargs
) -> SwaggerClient:
    """Generate a new ESI client.

    Args:
        token(esi.models.Token): used to access authenticated endpoints.
        datasource: Name of the ESI datasource to access.
        spec_file: Absolute path to a swagger spec file to load.
        version: Base ESI API version. Accepted values are 'legacy', 'latest',
        app_info_text: Text identifying the application using ESI which will be \
            included in the User-Agent header. Should contain name and version of the \
            application using ESI. e.g. `"my-app v1.0.0"`. \
            Note that spaces are used as delimiter.
        kwargs: Explicit resource versions to build, in the form Character='v4'. \
            Same values accepted as version.

    If a spec_file is specified, specific versioning is not available.
    Meaning the version and resource version kwargs are ignored in favour of the
    versions available in the spec_file.

    Returns:
        New ESI client
    """
    if app_settings.ESI_INFO_LOGGING_ENABLED:
        logger.info('Generating an ESI client...')

    client = RequestsClientPlus()
    user_agent = (
        str(app_info_text) if app_info_text else f"{__title__} v{__version__}"
    )
    if app_settings.ESI_USER_CONTACT_EMAIL:
        user_agent += f" {app_settings.ESI_USER_CONTACT_EMAIL}"

    client.user_agent = user_agent

    my_http_adapter = HTTPAdapter(
        pool_maxsize=app_settings.ESI_CONNECTION_POOL_MAXSIZE,
        max_retries=app_settings.ESI_CONNECTION_ERROR_MAX_RETRIES
    )
    client.session.mount('https://', my_http_adapter)

    if token or datasource:
        client.authenticator = TokenAuthenticator(token=token, datasource=datasource)

    api_version = version or app_settings.ESI_API_VERSION

    if spec_file:
        return read_spec(spec_file, http_client=client)
    else:
        spec = build_spec(api_version, http_client=client, **kwargs)
        return SwaggerClient(spec)


def minimize_spec(spec_dict, operations=None, resources=None):
    """
    Trims down a source spec dict to only the operations or resources indicated.
    :param spec_dict: The source spec dict to minimize.
    :type spec_dict: dict
    :param operations: A list of operation IDs to retain.
    :type operations: list of str
    :param resources: A list of resource names to retain.
    :type resources: list of str
    :return: Minimized swagger spec dict
    :rtype: dict
    """
    operations = operations or []
    resources = resources or []

    # keep the ugly overhead for now but only add paths we need
    minimized = {key: value for key, value in spec_dict.items() if key != 'paths'}
    minimized['paths'] = {}

    for path_name, path in spec_dict['paths'].items():
        for method, data in path.items():
            if (
                data['operationId'] in operations
                or any(tag in resources for tag in data['tags'])
            ):
                if path_name not in minimized['paths']:
                    minimized['paths'][path_name] = {}
                minimized['paths'][path_name][method] = data

    return minimized


class EsiClientProvider:
    """Class for providing a single ESI client instance for the whole app

    Args:
        datasource: Name of the ESI datasource to access.
        spec_file: Absolute path to a swagger spec file to load.
        version: Base ESI API version. Accepted values are 'legacy', 'latest',
        app_info_text: Text identifying the application using ESI which will be \
            included in the User-Agent header. Should contain name and version of \
            the application using ESI. e.g. `"my-app v1.0.0"`. \
            Note that spaces are used as delimiter.
        kwargs: Explicit resource versions to build, in the form Character='v4'. \
            Same values accepted as version.

    If a spec_file is specified, specific versioning is not available.
    Meaning the version and resource version kwargs are ignored in favour of the
    versions available in the spec_file.
    """

    _client = None

    def __init__(
        self,
        datasource=None,
        spec_file=None,
        version=None,
        app_info_text=None,
        **kwargs
    ):
        self._datasource = datasource
        self._spec_file = spec_file
        self._version = version
        self._app_text = app_info_text
        self._kwargs = kwargs

    @property
    def client(self):
        if self._client is None:
            self._client = esi_client_factory(
                datasource=self._datasource,
                spec_file=self._spec_file,
                version=self._version,
                app_info_text=self._app_text,
                **self._kwargs,
            )
        return self._client

    def __str__(self):
        return 'EsiClientProvider'



from datetime import datetime
from django.core.management.base import BaseCommand, CommandError

from django.db.migrations.recorder import MigrationRecorder
from django.utils import timezone

from esi.models import Token
from esi.errors import (
    TokenInvalidError,
    IncompleteResponseError,
    NotRefreshableTokenError,
)
from esi import app_settings
from oauthlib.oauth2.rfc6749.errors import (
    InvalidClientIdError,
    InvalidGrantError,
    InvalidTokenError,
)
from requests.auth import HTTPBasicAuth

from tqdm import tqdm

from requests_oauthlib import OAuth2Session

import pytz

EVE_SSOV1_END_DATE = pytz.UTC.localize(
    datetime(year=2021, month=11, day=1, hour=0, minute=0)
)


def _sso_v1_refresh(
    session: OAuth2Session, auth: HTTPBasicAuth, token: Token, message: str
):
    """
    SSOv1 Refresh.
    """
    try:
        _data = session.refresh_token(
            "https://login.eveonline.com/oauth/token",
            refresh_token=token.refresh_token,
            auth=auth,
        )
        token.access_token = _data["access_token"]
        token.refresh_token = _data["refresh_token"]
        token.sso_version = 1
        token.created = timezone.now()
        token.save()
        return (True, None)
    except (InvalidGrantError):
        return (
            False,
            (
                "ID:%s '%s' %s SSOv1 Refresh failed"
                " (InvalidGrant)" % (token.id, token.character_name, message)
            ),
        )
    except (InvalidTokenError, InvalidClientIdError):
        return (
            False,
            (
                "ID:%s '%s' %s SSOv1 Refresh Failed "
                "(InvalidToken, InvalidClientId)"
                % (token.id, token.character_name, message)
            ),
        )
    except Exception as e:
        return (
            False,
            (
                "ID:%s '%s' %s SSOv1 "
                "Refresh Failed (%s)" % (token.id, token.character_name, message, e)
            ),
        )


class Command(BaseCommand):
    help = "Attempt to Migrate all SSOv1 Tokens to SSOv2, and report failures."
    requires_migrations_checks = True
    requires_system_checks = True

    def add_arguments(self, parser):
        parser.add_argument(
            "--skip-v1-checks",
            action="store_true",
            help="Do not test on SSOv1 Endpoints, both before"
            " updating to SSOv2 and after a failure.",
        )
        parser.add_argument(
            "--purge",
            action="store_true",
            help="Purge Tokens that fail the SSOv2 Update",
        )
        parser.add_argument("-n", help="Number of Tokens to update", type=int)

    def handle(self, *args, **options):
        if timezone.now() > EVE_SSOV1_END_DATE:
            self.stdout.write(
                self.style.WARNING(
                    "WARNING: Conversion of SSO v1 tokens is not garenteered to work "
                    f"past {EVE_SSOV1_END_DATE.strftime('%x')}."
                )
            )
        use_v1 = not options.get("skip_v1_checks", False)
        purge = options.get("purge", False)
        batch = options.get("n", False)

        migration_10 = MigrationRecorder.Migration.objects.filter(
            app="esi", name="0010_set_new_tokens_to_sso_v2"
        ).exists()

        if not migration_10:
            raise CommandError("Run migrations first and try again!")
        else:
            self.stdout.write("\n\nMigrations up to date. Proceeding to updates!")

        if not use_v1:
            self.stdout.write("\n\nSkipping SSOv1 pre/post error Verification!")

        # lets reuse our own session and auth
        session = OAuth2Session(app_settings.ESI_SSO_CLIENT_ID)
        auth = HTTPBasicAuth(
            app_settings.ESI_SSO_CLIENT_ID, app_settings.ESI_SSO_CLIENT_SECRET
        )

        # only do tokens that are SSOv1
        tokens = Token.objects.filter(sso_version=1)
        if batch:
            tokens = tokens[:batch]
        total = tokens.count()

        if total > 0:
            self.stdout.write("There are %s Tokens to update." % (total))
        else:
            return self.stdout.write("There are no Tokens to update.")

        failures = []
        # Nice Progress bar as this may take a while.
        with tqdm(total=total) as t:
            for token in tokens:
                if use_v1:
                    result, message = _sso_v1_refresh(session, auth, token, "Initial")
                    if not result:
                        failures.append(message)
                        if purge:
                            token.delete()
                        t.update(1)
                        continue
                try:
                    token.refresh(session=session, auth=auth)
                except (TokenInvalidError, IncompleteResponseError):
                    if use_v1:
                        result, messaage = _sso_v1_refresh(
                            session, auth, token, "Post v2 Failure"
                        )
                        if not result:
                            failures.append(messaage)
                            if purge:
                                token.delete()
                        else:
                            failures.append(
                                "ID:%s '%s' SSOv2 Migration failed"
                                " (SSOv1 Re-Verification Ok)"
                                % (token.id, token.character_name)
                            )

                    else:
                        failures.append(
                            "ID:%s '%s' SSOv2 Migration failed"
                            " (TokenInvalidError, "
                            "IncompleteResponseError)"
                            % (token.id, token.character_name)
                        )
                        if purge:
                            token.delete()
                except NotRefreshableTokenError:
                    failures.append(
                        "ID:%s '%s' SSOv2 Migration failed"
                        " (NotRefreshableTokenError)" % (token.id, token.character_name)
                    )
                    if purge:
                        token.delete()

                t.update(1)

        self.stdout.write("Completed Updates!")
        self.stdout.write("%s Failures!" % (len(failures)))
        self.stdout.write("\n".join(failures))



from datetime import timedelta
import logging
from typing import Union

import requests
from requests_oauthlib import OAuth2Session

from django.db import models
from django.utils import timezone

from .errors import TokenError, IncompleteResponseError
from . import app_settings

from jose import jwt
from jose.exceptions import ExpiredSignatureError, JWTError, JWTClaimsError

logger = logging.getLogger(__name__)


def _process_scopes(scopes):
    if scopes is None:
        # support filtering by no scopes with None passed
        scopes = []
    if not isinstance(scopes, models.QuerySet) and len(scopes) == 1:
        # support a single space-delimited string inside a list because :users:
        scopes = scopes[0]
    # support space-delimited string scopes or lists
    if isinstance(scopes, str):
        scopes = set(scopes.split())
    return {str(s) for s in scopes}


class TokenQueryset(models.QuerySet):
    def get_expired(self) -> models.QuerySet:
        """Get all tokens which have expired.

        Returns:
            All expired tokens.
        """
        max_age = \
            timezone.now() - timedelta(seconds=app_settings.ESI_TOKEN_VALID_DURATION)
        return self.filter(created__lte=max_age)

    def bulk_refresh(self) -> models.QuerySet:
        """Refresh all refreshable tokens in the queryset and delete any expired token
        that fails to refresh or can not be refreshed.

        Excludes tokens for which the refresh was incomplete for other reasons.

        Returns:
            All refreshed tokens
        """
        session = OAuth2Session(app_settings.ESI_SSO_CLIENT_ID)
        auth = requests.auth.HTTPBasicAuth(
            app_settings.ESI_SSO_CLIENT_ID, app_settings.ESI_SSO_CLIENT_SECRET
        )
        incomplete = []
        for model in self.filter(refresh_token__isnull=False):
            try:
                model.refresh(session=session, auth=auth)
                logging.debug("Successfully refreshed %r", model)
            except TokenError:
                logger.info("Refresh failed for %r. Deleting.", model)
                model.delete()
            except IncompleteResponseError:
                incomplete.append(model.pk)
        self.filter(refresh_token__isnull=True).get_expired().delete()
        return self.exclude(pk__in=incomplete)

    def require_valid(self) -> models.QuerySet:
        """Ensure all tokens are still valid and attempt to refresh any which are expired

        Deletes those which fail to refresh or cannot be refreshed.

        Returns:
            All tokens which are still valid.
        """
        expired = self.get_expired()
        valid = self.exclude(pk__in=expired)
        valid_expired = expired.bulk_refresh()
        return valid_expired | valid

    def require_scopes(self, scope_string: Union[str, list]) -> models.QuerySet:
        """Filter tokens which have at least a subset of given scopes.

        Args:
            scope_string: The required scopes.

        Returns:
            Tokens which have all requested scopes.
        """
        scopes = _process_scopes(scope_string)
        if not scopes:
            # asking for tokens with no scopes
            return self.filter(scopes__isnull=True)
        from .models import Scope
        scope_pks = Scope.objects.filter(name__in=scopes).values_list('pk', flat=True)
        if not len(scopes) == len(scope_pks):
            # there's a scope we don't recognize, so we can't have any tokens for it
            return self.none()
        tokens = self.all()
        for pk in scope_pks:
            tokens = tokens.filter(scopes__pk=pk)
        return tokens

    def require_scopes_exact(self, scope_string: Union[str, list]) -> models.QuerySet:
        """Filter tokens which exactly have the given scopes.

        Args:
            scope_string: The required scopes.

        Returns:
            Tokens which have all requested scopes.
        """
        num_scopes = len(_process_scopes(scope_string))
        scopes_qs = self\
            .annotate(models.Count('scopes'))\
            .require_scopes(scope_string)\
            .filter(scopes__count=num_scopes)\
            .values('pk', 'scopes__id')
        pks = [v['pk'] for v in scopes_qs]
        return self.filter(pk__in=pks)

    def equivalent_to(self, token) -> models.QuerySet:
        """Fetch all tokens which match the character and scopes of given reference token

        Args:
            token: :class:`esi.models.Token` reference token
        """
        return self\
            .filter(character_id=token.character_id)\
            .require_scopes_exact(token.scopes.all())\
            .filter(models.Q(user=token.user) | models.Q(user__isnull=True))\
            .exclude(pk=token.pk)


class TokenManager(models.Manager):
    def get_queryset(self):
        """
        Replace base queryset model with custom TokenQueryset
        :rtype: :class:`esi.managers.TokenQueryset`
        """
        return TokenQueryset(self.model, using=self._db)

    @staticmethod
    def _decode_jwt(jwt_token: dict, jwk_set: dict, issuer: str):
        """
        Helper function to decide the JWT access token supplied by EVE SSO
        """
        logger.debug("Start Decode")
        token_data = jwt.decode(
            jwt_token,
            jwk_set,
            algorithms=jwk_set["alg"],
            issuer=issuer
        )
        token_detail = token_data.get("sub", None).split(":")
        token_data['character_id'] = int(token_detail[2])
        token_data['token_type'] = token_detail[0].lower()
        logger.debug(token_data)
        return token_data

    @staticmethod
    def validate_access_token(token: str):
        """
        Validate a JWT token retrieved from the EVE SSO.
        :param token: A JWT token originating from the EVE SSO v2
        :return: :class:`dict` The contents of the validated JWT token if
            there are no validation errors
        """

        res = requests.get(app_settings.ESI_TOKEN_JWK_SET_URL)
        res.raise_for_status()

        data = res.json()

        try:
            jwk_sets = data["keys"]
        except KeyError as e:
            logger.warning("Something went wrong when retrieving the JWK set. "
                           "The returned payload did not have the expected key"
                           "{}. \nPayload returned from the SSO looks like: "
                           "{}".format(e, data))
            return None

        jwk_set = next(item for item in jwk_sets if item["alg"] == "RS256")

        try:
            return TokenManager._decode_jwt(
                token,
                jwk_set,
                "login.eveonline.com"
            )
        # TODO Raise the errors to be handled in the lib
        except ExpiredSignatureError:
            logger.warning("The JWT token has expired: {}")
            return None
        except JWTError as e:
            logger.warning(f"The JWT signature was invalid: {str(e)}")
            return None
        except JWTClaimsError:
            try:
                return TokenManager._decode_jwt(
                    token,
                    jwk_set,
                    "https://login.eveonline.com"
                )
            except JWTClaimsError as e:
                logger.warning("The issuer claim was not from "
                               "login.eveonline.com or "
                               "https://login.eveonline.com: "
                               "{}".format(str(e)))
                return None

    def create_from_code(self, code, user=None):
        """
        Perform OAuth code exchange to retrieve a token.
        :param code: OAuth grant code.
        :param user: User who will own token.
        :return: :class:`esi.models.Token`
        """

        # perform code exchange
        logger.debug("Creating new token from code %s", code[:-5])
        oauth = OAuth2Session(
            app_settings.ESI_SSO_CLIENT_ID,
            redirect_uri=app_settings.ESI_SSO_CALLBACK_URL
        )
        token = oauth.fetch_token(
            app_settings.ESI_TOKEN_URL,
            client_secret=app_settings.ESI_SSO_CLIENT_SECRET,
            code=code
        )

        token_data = TokenManager.validate_access_token(token.get('access_token', None))

        # translate returned data to a model
        model = self.create(
            character_id=token_data['character_id'],
            character_name=token_data['name'],
            character_owner_hash=token_data['owner'],
            access_token=token['access_token'],
            refresh_token=token['refresh_token'],
            token_type=token_data['token_type'],
            user=user,
        )

        # parse scopes
        if 'scp' in token_data:
            from esi.models import Scope

            # if a single scope is supplied its a string... recast to list
            if isinstance(token_data['scp'], str):
                token_data['scp'] = [token_data['scp']]

            for s in token_data['scp']:
                try:
                    scope = Scope.objects.get(name=s)
                    model.scopes.add(scope)
                except Scope.DoesNotExist:
                    # This scope isn't included in a data migration.
                    # Create a placeholder until it updates.
                    try:
                        help_text = s.split('.')[1].replace('_', ' ').capitalize()
                    except IndexError:
                        # Unusual scope name, missing periods.
                        help_text = s.replace('_', ' ').capitalize()
                    scope = Scope.objects.create(name=s, help_text=help_text)
                    model.scopes.add(scope)
            logger.debug("Added %d scopes to new token.", model.scopes.all().count())

        if not app_settings.ESI_ALWAYS_CREATE_TOKEN:
            # see if we already have a token for this character and scope combination
            # if so, we don't need a new one
            queryset = self.get_queryset().equivalent_to(model)
            if queryset.exists():
                logger.debug(
                    "Identified %d tokens equivalent to new token. "
                    "Updating access and refresh tokens.",
                    queryset.count()
                )
                queryset.update(
                    access_token=model.access_token,
                    refresh_token=model.refresh_token,
                    created=model.created,
                )
                if queryset.filter(user=model.user).exists():
                    logger.debug(
                        "Equivalent token with same user exists. Deleting new token."
                    )
                    model.delete()
                    model = queryset.filter(user=model.user)[0]  # pick one at random

        logger.debug("Successfully created %r for user %s", model, user)
        return model

    def create_from_request(self, request):
        """
        Generate a token from the OAuth callback request. Must contain 'code' in GET.
        :param request: OAuth callback request.
        :return: :class:`esi.models.Token`
        """
        logger.debug(
            "Creating new token for %s session %s",
            request.user,
            request.session.session_key[:5]
        )
        code = request.GET.get('code')
        # attach a user during creation for some functionality in a post_save created
        # receiver I'm working on elsewhere
        model = self.create_from_code(
            code, user=request.user if request.user.is_authenticated else None
        )
        return model



# Generated by Django 3.1.1 on 2021-10-08 10:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0009_set_old_tokens_to_sso_v1'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='sso_version',
            field=models.IntegerField(default=2, help_text='EVE SSO Version.'),
        ),
    ]



# Generated by Django 3.1.3 on 2020-11-09 23:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0007_fix_mysql_8_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='refresh_token',
            field=models.TextField(editable=False, help_text='A re-usable token to generate new access tokens upon expiry.', null=True),
        ),
    ]



from django.apps import AppConfig


class EsiConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'esi'
    verbose_name = 'EVE Swagger Interface (SSO v2)'

    def ready(self):
        super().ready()
        from esi import checks  # noqa



import datetime
import re
import logging

from bravado.client import SwaggerClient
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session
from oauthlib.oauth2.rfc6749.errors import (
    InvalidClientError,
    InvalidClientIdError,
    InvalidGrantError,
    InvalidTokenError,
    MissingTokenError,
)

from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.db import models
from django.utils import timezone

from . import app_settings
from .clients import esi_client_factory
from .managers import TokenManager
from .errors import (
    IncompleteResponseError,
    NotRefreshableTokenError,
    TokenExpiredError,
    TokenInvalidError,
)


logger = logging.getLogger(__name__)


class Scope(models.Model):
    """
    Represents an access scope granted by SSO.
    """
    name = models.CharField(
        max_length=100, unique=True, help_text="The official EVE name for the scope."
    )
    help_text = models.TextField(help_text="The official EVE description of the scope.")

    @property
    def friendly_name(self):
        return self._friendly_name(self.name)

    @classmethod
    def _friendly_name(cls, name):
        try:
            return re.sub('_', ' ', name.split('.')[1]).strip()
        except IndexError:
            return name

    def __str__(self):
        return self.name


class Token(models.Model):
    """EVE Swagger Interface Access Token

    Contains information about the authenticating character
    and scopes granted to this token.
    Contains the access token required for ESI authentication as well as refreshing.
    """

    created = models.DateTimeField(auto_now_add=True)
    access_token = models.TextField(
        help_text="The access token granted by SSO.",
        editable=False
    )
    refresh_token = models.TextField(
        null=True,  # refresh tokens returned from SSO can be null
        help_text="A re-usable token to generate new access tokens upon expiry.",
        editable=False
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="The user to whom this token belongs."
    )
    character_id = models.IntegerField(
        help_text="The ID of the EVE character who authenticated by SSO."
    )
    character_name = models.CharField(
        max_length=100,
        help_text="The name of the EVE character who authenticated by SSO."
    )
    token_type = models.CharField(
        max_length=100,
        choices=(('Character', 'Character'), ('Corporation', 'Corporation'),),
        default='Character',
        help_text="The applicable range of the token."
    )
    character_owner_hash = models.CharField(
        max_length=254,
        help_text=(
            "The unique string identifying this character and its owning EVE "
            "account. Changes if the owning account changes."
        )
    )
    scopes = models.ManyToManyField(
        Scope, blank=True, help_text="The access scopes granted by this token."
    )
    sso_version = models.IntegerField(
        help_text="EVE SSO Version.",
        default=2
    )

    objects = TokenManager()

    def __str__(self):
        return "{} - {}".format(
            self.character_name, ", ".join(sorted(s.name for s in self.scopes.all()))
        )

    def __repr__(self):
        return "<{}(id={}): {}, {}>".format(
            self.__class__.__name__,
            self.pk,
            self.character_id,
            self.character_name,
        )

    @property
    def can_refresh(self) -> bool:
        """Determine if this token can be refreshed upon expiry."""
        return bool(self.refresh_token)

    @property
    def expires(self) -> datetime.datetime:
        """Determines when this token expires.

        Returns:
            Date & time when this token expires
        """
        return (
            self.created
            + datetime.timedelta(seconds=app_settings.ESI_TOKEN_VALID_DURATION)
        )

    @property
    def expired(self) -> bool:
        """Determines if this token has expired."""
        return self.expires < timezone.now()

    def valid_access_token(self) -> str:
        """Refresh and return access token to be used in an authed ESI call.

        Example:
            .. code-block:: python

                # fetch medals for a character
                medals = esi.client.Character.get_characters_character_id_medals(
                    # required parameter for endpoint
                    character_id = token.character_id,
                    # provide a valid access token, which will be refreshed if required
                    token = token.valid_access_token()
                ).results()

        Returns:
           Valid access token

        Raises:
            TokenExpiredError: When token can not be refreshed
        """
        if self.expired:
            if self.can_refresh:
                self.refresh()
            else:
                raise TokenExpiredError()
        return self.access_token

    def refresh(
        self, session: OAuth2Session = None, auth: HTTPBasicAuth = None
    ) -> None:
        """Refresh this token.

        Args:
            session: session for refreshing token with
            auth: ESI authentication
        """
        logger.debug("Attempting refresh of %r", self)
        if self.can_refresh:
            if not session:
                session = OAuth2Session(app_settings.ESI_SSO_CLIENT_ID)
            if not auth:
                auth = HTTPBasicAuth(
                    app_settings.ESI_SSO_CLIENT_ID, app_settings.ESI_SSO_CLIENT_SECRET
                )
            try:
                token = session.refresh_token(
                    app_settings.ESI_TOKEN_URL,
                    refresh_token=self.refresh_token,
                    auth=auth
                )
                logger.debug("Retrieved new token from SSO servers.")
                # logger.debug(token)
                token_data = TokenManager.validate_access_token(token['access_token'])

                # TODO verify token properly
                if token_data is not None:
                    if self.character_owner_hash != token_data['owner']:
                        logger.warning("Invalid Owner")
                        raise InvalidTokenError("Ownership Changed! Revoke me!")

                self.access_token = token['access_token']
                self.refresh_token = token['refresh_token']
                self.sso_version = 2  # we will never be ssov1 again
                self.created = timezone.now()
                self.save()
                logger.debug("Successfully refreshed %r", self)
            except (InvalidGrantError) as e:
                # this token is gone forever
                logger.error("Refresh impossible for %r: %r", self, e)
                raise TokenInvalidError()
            except (InvalidTokenError, InvalidClientIdError) as e:
                # these may be recoverable?
                logger.warning("Refresh failed for %r: %r", self, e)
                raise TokenInvalidError()
            except MissingTokenError as e:
                logger.info("Refresh failed for %r: %r", self, e)
                raise IncompleteResponseError()
            except InvalidClientError:
                logger.debug(
                    "ESI client ID and secret rejected by remote. Cannot refresh."
                )
                raise ImproperlyConfigured(
                    'Verify ESI_SSO_CLIENT_ID and ESI_SSO_CLIENT_SECRET settings.'
                )
        else:
            logger.debug("Not a refreshable token.")
            raise NotRefreshableTokenError()

    def get_esi_client(self, **kwargs) -> SwaggerClient:
        """Creates an authenticated ESI client with this token.

        Args:
            **kwargs: Extra spec versioning as per \
                :class:`esi.clients.esi_client_factory`

        Returns:
            New ESI client
        """
        return esi_client_factory(token=self, **kwargs)

    @classmethod
    def get_token_data(cls, access_token):
        return TokenManager.validate_access_token(access_token)

    # unused?
    def update_token_data(self, commit=True):
        logger.debug("Updating token data for %r", self)
        if self.expired:
            if self.can_refresh:
                self.refresh()
            else:
                raise TokenExpiredError()
        token_data = self.get_token_data(self.access_token)
        logger.debug(token_data)
        self.character_id = token_data['character_id']
        self.character_name = token_data['name']
        self.character_owner_hash = token_data['owner']
        self.token_type = token_data['token_type']
        logger.debug("Successfully updated token data.")
        if commit:
            self.save()

    @classmethod
    def get_token(cls, character_id: int, scopes: list) -> "Token":
        """Helper method to get a token for a specific character with specific scopes.

        Args:
            character_id: Character to filter on.
            scopes: array of ESI scope strings to search for.

        Returns:
            Matching token or `False` when token is not found
        """
        token = (
            Token.objects
            .filter(character_id=character_id)
            .require_scopes(scopes)
            .first()
        )
        if token:
            return token
        else:
            return False


class CallbackRedirect(models.Model):
    """
    Records the intended destination for the SSO callback.
    Used to internally redirect SSO callbacks.
    """
    session_key = models.CharField(
        max_length=254,
        unique=True,
        help_text="Session key identifying the session this redirect was created for."
    )
    url = models.TextField(
        default="/",
        help_text="The internal URL to redirect this callback towards."
    )
    state = models.CharField(
        max_length=128,
        help_text="OAuth2 state string representing this session."
    )
    created = models.DateTimeField(auto_now_add=True)
    token = models.ForeignKey(
        Token,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text=(
            "Token generated by a completed code exchange "
            "from callback processing."
        )
    )

    def __str__(self):
        return f"{self.session_key}: {self.url}"

    def __repr__(self):
        return "<{}(pk={}): {} to {}>".format(
            self.__class__.__name__, self.pk,
            self.session_key, self.url
        )



import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
BASE_DIR = os.path.dirname(PROJECT_DIR)

SECRET_KEY = "wow I'm a really bad default secret key"

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "debug_toolbar.middleware.DebugToolbarMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "tests.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = 'mysite.wsgi.application'

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": str(os.path.join(BASE_DIR, "mysite.sqlite3")),
    },
}

# Password validation
# https://docs.djangoproject.com/en/2.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME':
        'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = '/static/'


INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "esi",
]


DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

# esi
ESI_API_URL = "https://esi.evetech.net/"
ESI_SSO_CLIENT_ID = "test-dummy"
ESI_SSO_CLIENT_SECRET = "test-dummy"
ESI_SSO_CALLBACK_URL = "http://localhost:8000"

# timezone.now() needs to be tz_aware
USE_TZ = True



# Generated by AaronKable on 2020-10-03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0006_remove_url_length_limit'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='refresh_token',
            field=models.TextField(editable=False, help_text='A re-usable token to generate new access tokens upon expiry.'),
        ),
    ]



# Generated by Django 3.1.1 on 2021-10-08 10:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0008_nullable_refresh_token'),
    ]

    operations = [
        migrations.AddField(
            model_name='token',
            name='sso_version',
            field=models.IntegerField(default=1, help_text='EVE SSO Version.'),
        ),
    ]
